from flask import Flask
from flask import render_template, request, abort

import git
import waitress

import shutil
from pathlib import Path
import time
import datetime
import json
from typing import Tuple
from collections import defaultdict

app = Flask(__name__)


TMP_PATH = Path('/tmp/coldbrew/')
base = Path.cwd()


def create_bare_repo(repo_path: Path):
    """
    Creates a bare repo at the given repo_path
    :param repo_path: The path to the bare repo (usually ending with something.git/). This path will be created
    :return:
    """
    assert not repo_path.is_dir(), f"Directory {repo_path} does already exist. Cannot init a new repo here"
    bare_repo = git.Repo.init(repo_path, bare=True)
    assert bare_repo.bare


def create_meta(path: Path, hash: str, description: str, prepare_timestamp: int):
    """
    Creates the directory path and the 3 files commit-hash.txt, description.txt, prepare-timestamp.txt within that directory
    :param path: Path to the meta directory. Cannot already exist
    :param hash: Commit hash, will be saved in commit-hash.txt
    :param description: Commit description, will be saved in description.txt
    :param prepare_timestamp: The timestamp when preparation was started as ms since epoch (created by int(time.time()))
    :return:
    """
    path.mkdir()
    with open(path / 'commit-hash.txt', 'a') as f:
        f.write(hash)
    with open(path / 'description.txt', 'a') as f:
        f.write(description)
    with open(path / 'prepare-timestamp.txt', 'a') as f:
        f.write(str(prepare_timestamp))


def checkout_commit(target_path: Path, commit_hash: str) -> Tuple[str, str]:
    """
    Checkout a commit given by commit_hash into target_dir.
    As this is designed to work with bare repos, it will actually perform a clone() and then remove the .git/ directory from target_dir
    :param target_path: Path to target directory
    :param commit_hash: The hexsha hash uniquely identifying the desired commit
    :return: The commit's description (part of commit message without <<>>), the commit's group (content within << and >>)
    """
    try:
        repo = git.Repo(base / 'coldbrew.git')
    except git.InvalidGitRepositoryError as e:
        raise AssertionError(f"{base} does not contain a valid git repo (Bare repo expected in {base / 'coldbrew.git'}).")

    if commit_hash not in [c.hexsha for c in repo.iter_commits(rev=f'coldbrew')]:
        raise AssertionError(f"The supplied commit hash is not valid. Check if the last push was successful.")

    # Save the current head state
    try:
        current_head = repo.head.reference
    except:
        pass

    target_path_str = str(target_path)
    if not target_path_str[-1] == '/':  # GitPython clone requires a / at the end!
        target_path_str = f'{target_path_str}/'

    # Checkout code from the requested commit into a /tmp directory
    target_commit = repo.commit(commit_hash)
    repo.head.reference = target_commit
    description = target_commit.message.strip().split(' <<')[0]
    if '<<' in target_commit.message and '>>' in target_commit.message:
        group = target_commit.message.strip().split(' <<')[1].split('>>')[0]
    else:
        group = ''
    repo.clone(path=target_path_str)
    shutil.rmtree(target_path / '.git', ignore_errors=True)

    # Restore the repo head state
    try:
        repo.head.reference = current_head
    except:
        pass

    return description, group


def prepare_directories(commit_hash: str) -> dict:
    """
    Prepare the directories for an experiment using code from commit given by commit_hash
    :param commit_hash: The hexsha hash uniquely identifying the desired commit
    :return: Dict containing code and results paths in str format
    """
    # Generate experiment ID as a combination of short commit hash and prepare_timestamp
    prepare_timestamp = time.time()
    id = f"{commit_hash[:7]}-{int(prepare_timestamp)}"

    # Code goes into /tmp
    code_path = Path('/tmp/coldbrew/') / id

    # Checkout this commit into the defined code directory
    description, group = checkout_commit(code_path, commit_hash)

    # Create result directory structure
    results_path = base / group / id
    results_path.mkdir(parents=True)

    # Write the meta information
    create_meta(results_path / 'meta', commit_hash, description, int(prepare_timestamp))

    # Return the command to be run including all generated directory names
    return {
        'code': str(code_path),
        'results': str(results_path)
    }


def get_commits(max_num_entries=25) -> list:
    """
    Get the n most recent commits from coldbrew.git (branch coldbrew) in the current directory
    :param max_num_entries: Max number of commits to retrieve
    :return: List of commits (hash, short_hash, description, commit_timestamp), sorted by commit_timestamp
    """
    repo = git.Repo(base / 'coldbrew.git')
    if 'coldbrew' not in repo.branches:
        return []  # If repo has not yet been pushed to, show empty list
    commits = [
        {
            'hash': commit.hexsha,
            'short_hash': commit.hexsha[:7],
            'description': commit.message.strip().split(' <<')[0],
            'group': '' if not ('<<' in commit.message and '>>' in commit.message) else (commit.message.strip().split(' <<')[1].split('>>')[0]),
            'commit_timestamp': commit.committed_date,
        }
        for commit in list(repo.iter_commits(rev=f'coldbrew'))[:max_num_entries]
    ]
    return sorted(commits, key=lambda c: c['commit_timestamp'], reverse=True)


def get_experiments() -> Tuple[list, str]:
    """
    Get all already prepared experiments: All subdirectories of base with a dash at position 7
    :return: List of experiments and json string of dict mapping experiment ids to experiments
    """
    experiments = []
    group_dates = {}
    # Recursively go over all subdirectories of base
    for path in base.rglob(''):
        # Make sure path actually is a results path
        if len(path.name) <= 7 or not path.name[7] == '-':
            continue
        # Get metadata
        with open(path / 'meta' / 'commit-hash.txt', 'r') as f:
            hash = f.read()
        with open(path / 'meta' / 'description.txt', 'r') as f:
            description = f.read()
        with open(path / 'meta' / 'prepare-timestamp.txt', 'r') as f:
            start = float(f.read())
        # Generate group name: Not in subdir => Empty group; In multiple hierarchies => Repace / with -
        group_parts = path.relative_to(base).parts
        if len(group_parts) == 1:
            group = ''
        else:
            group = '-'.join(group_parts[:-1])
        # Get latest date of experiments within each group; used for sorting the table in the webinterface
        if group not in group_dates or group_dates[group] < start:
            group_dates[group] = start
        # Get code and results paths
        path_code = f'/tmp/coldbrew/{hash[:7]}-{int(start)}'
        path_results = str(path)
        # Compose experiments
        experiment = {
            'id': f'{hash[:7]}-{int(start)}',
            'short_id': hash[:7],
            'hash': hash,
            'description': description,
            'prepare_datetime': datetime.datetime.fromtimestamp(start).strftime('%Y-%m-%d %H:%M:%S'),
            'path_code': path_code,
            'path_results': path_results,
            'group': group,
            'group_date': None,
            'cmd_strings': []
        }
        # Templating
        if (Path(path_code) / '.coldbrew').is_file():
            with open(Path(path_code) / '.coldbrew', 'r') as f:
                cmd_string_tmpls = [l for l in (line.strip() for line in f) if l]
            experiment['cmd_strings'].extend([tmpl.format_map(defaultdict(str, code_path=path_code, results_path=path_results)) for tmpl in cmd_string_tmpls])
        if (base / '.coldbrew').is_file():
            with open(base / '.coldbrew', 'r') as f:
                cmd_string_tmpls = [l for l in (line.strip() for line in f) if l]
            experiment['cmd_strings'].extend([tmpl.format_map(defaultdict(str, code_path=path_code, results_path=path_results)) for tmpl in cmd_string_tmpls])
        # Append experiment
        experiments.append(experiment)
        # Insert max group timestamps
        for experiment in experiments:
            experiment['group_date'] = group_dates[experiment['group']]

    # Need to return the json dict for the web interface javascript implementation
    return experiments, json.dumps({exp['id']: exp for exp in experiments})


# FLASK ROUTES
@app.route('/code/', defaults={'req_path': '/'})
@app.route('/code/<path:req_path>')
def code(req_path):
    if '/..' in req_path:
        return abort(404)

    abs_path = TMP_PATH / req_path

    experiment_id = req_path.split('/')[0]
    results_dir = None
    for path in base.rglob(''):
        if experiment_id == path.name:
            results_dir = path.absolute()
    if results_dir is None:
        return abort(404)
    with open(results_dir / 'meta' / 'description.txt', 'r') as f:
        description = f.read()
    with open(results_dir / 'meta' / 'commit-hash.txt', 'r') as f:
        commit_hash = f.read()

    if abs_path.is_file():
        with open(abs_path, 'r') as f:
            content = f.read()
        if not (TMP_PATH / req_path.split('/')[0]).is_dir():
            checkout_commit(TMP_PATH / req_path.split('/')[0], commit_hash)
        print(f"[INFO] Showing content of file {abs_path}")
        return render_template('file.tmpl',
                               parts=zip(abs_path.relative_to(TMP_PATH).parents, abs_path.relative_to(TMP_PATH).parts),
                               current_path=req_path, description=description, content=content)
    elif abs_path.is_dir():
        with open(results_dir / 'meta' / 'description.txt', 'r') as f:
            description = f.read()
        files = zip([path.relative_to(TMP_PATH) for path in abs_path.iterdir()],
                    [path.relative_to(abs_path) for path in abs_path.iterdir()])
        print(f"[INFO] Showing content of directory {abs_path}")
        return render_template('directory.tmpl',
                               parts=zip(abs_path.relative_to(TMP_PATH).parents, abs_path.relative_to(TMP_PATH).parts),
                               current_path=req_path, description=description, files=files)
    else:
        return abort(404)


@app.route('/delete', methods=['POST'])
def delete():
    experiment_id = request.get_data().decode("utf-8")
    results_dir = None
    for path in base.rglob(''):
        if experiment_id == path.name:
            results_dir = path.absolute()
    if results_dir is None:
        return ''
    code_dir = Path('/tmp/coldbrew/') / experiment_id

    if results_dir.is_dir():
        shutil.rmtree(results_dir)
    if code_dir.is_dir():
        shutil.rmtree(code_dir)

    print(f"[INFO] Deleted experiment {experiment_id} (directories {code_dir} and {results_dir})")
    return ''


@app.route('/prepare', methods=['POST'])
def prepare():
    commit_hash = request.get_data().decode("utf-8")
    try:
        paths = prepare_directories(commit_hash)
    except AssertionError as e:
        msg = f"[Error] {str(e)}"
        print(msg)
        return msg

    print(f"[INFO] Prepared directories {paths['code']} and {paths['results']}")
    return ''


@app.route('/')
def index():
    experiments, experiments_json = get_experiments()
    return render_template('experiments.tmpl', experiments=experiments, experiments_json=experiments_json, commits=get_commits())


def run(base_path: Path, host: str, port: int, development=False):
    """
    Run the coldbrew flask app either in development mode or behind a waitress server for production, depending on the development flag
    :param base_path: Path containing results subdirectories and coldbrew.git/
    :param host: Host to serve from
    :param port: Port to serve from
    :param development: If true, run flask server directly; otherwise, run behind waitress server
    :return:
    """
    global base
    base = base_path

    if development:
        app.run(host, port)
    else:
        waitress.serve(app, host=host, port=port)


if __name__ == '__main__':
    raise NotImplementedError("This script is not meant to be run directly")
